//
//  ASPageLayoutAccessory.m
//  ASPageLayoutAccessory
//
//  Created by 昭 佐藤 on 12/02/06.
//  Copyright (c) 2014年 SatoAkira. All rights reserved.
//

#import "ASPageLayoutAccessory.h"
#import <ASForm/ASForm.h>

@interface ASPageLayoutAccessory ()

@property	BOOL _verticallyCentered,_horizontallyCentered,_headerAndFooter,headerAndFooterEnabled;
@property	BOOL upsideDown; // NSTextViewを描画する際topMarginが下にとられる。従ってこのような場合には、upsideDownをYESに設定すると使いやすくなる。 //
@property	CGFloat _leftMargin,_rightMargin,_topMargin,_bottomMargin,imageableAreaOriginX,imageableAreaOriginY,leftSpace,bottomSpace,rightSpace,topSpace,leftSpaceSumTotal,bottomSpaceSumTotal,rightSpaceSumTotal,topSpaceSumTotal;
@property	NSInteger _verticalPagination,_horizontalPagination;
@property IBOutlet NSTextField *leftMarginField,*rightMarginField,*topMarginField,*bottomMarginField;
@property IBOutlet ASForm *originView,*widthView,*heightView;
@property IBOutlet NSButton *headerAndFooterSwitch;
@property IBOutlet NSPopUpButton *unitsPopUp;
@property IBOutlet NSMatrix *centeredMatrix;

@end

static NSString *PageLayoutLeftMarginKey = @"leftMargin";
static NSString *PageLayoutRightMarginKey = @"rightMargin";
static NSString *PageLayoutTopMarginKey = @"topMargin";
static NSString *PageLayoutBottomMarginKey = @"bottomMargin";
static NSString *PageLayoutHeaderAndFooterKey = @"headerAndFooter";
static NSString *PageLayoutVerticallyCenterdKey = @"verticallyCentered";
static NSString *PageLayoutHorizontallyCenterdKey = @"horizontallyCentered";
static NSString *PageLayoutVerticalPaginationKey = @"verticalPagination";
static NSString *PageLayoutHorizontalPaginationKey = @"horizontalPagination";
static NSString *PageLayoutImageableAreaOriginXKey = @"imageableAreaOriginX";
static NSString *PageLayoutImageableAreaOriginYKey = @"imageableAreaOriginY";
static NSString *PageLayoutLeftSpaceKey = @"leftSpace";
static NSString *PageLayoutBottomSpaceKey = @"bottomSpace";
static NSString *PageLayoutRightSpaceKey = @"rightSpace";
static NSString *PageLayoutTopSpaceKey = @"topSpace";
NSString *PageLayoutUpsideDownKey = @"upsideDown";
NSString *PageLayoutNibName = @"ASPageLayoutAccessory";

@implementation ASPageLayoutAccessory

- (instancetype)initWithNibName:(NSString *)nibNameOrNil bundle:(NSBundle *)nibBundleOrNil
{
	NSBundle *bundle = (nil != nibBundleOrNil) ? nibBundleOrNil : [NSBundle bundleForClass:[ASPageLayoutAccessory class]];
	NSString *nibName = (nil != nibNameOrNil) ? nibNameOrNil : PageLayoutNibName;
	self.headerAndFooterEnabled = YES;
	self.upsideDown = NO; // NSTextViewを描画する際topMarginが下にとられる。従ってこのような場合には、upsideDownをYESに設定すると使いやすくなる。 //
    return [super initWithNibName:nibName bundle:bundle];
}
- (void)dealloc {
	[super dealloc];
}

- (NSDictionary *)pointsPerUnit {
	return [NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithDouble:28.35],@"cm",[NSNumber numberWithDouble:72.0],@"inch",[NSNumber numberWithDouble:12.0],@"pica",[NSNumber numberWithDouble:1.0],@"point",nil];
}

- (void)awakeFromNib {
	// setRepresentedObjectより前に呼ばれる。width=402のviewの中へwidth=402のboxを作ればぎりぎりいっぱいにboxが表示される。 //
	BOOL usesMetric = [[[NSLocale currentLocale] objectForKey:NSLocaleUsesMetricSystem] boolValue];
	NSString *defaultUnit = (YES == usesMetric) ? @"Metric" : [[NSLocale currentLocale] objectForKey:NSLocaleMeasurementSystem];

	[self.leftMarginField bind:NSValueBinding toObject:self withKeyPath:PageLayoutLeftMarginKey options:[NSDictionary dictionary]];
	[self.rightMarginField bind:NSValueBinding toObject:self withKeyPath:PageLayoutRightMarginKey options:[NSDictionary dictionary]];
	[self.topMarginField bind:NSValueBinding toObject:self withKeyPath:PageLayoutTopMarginKey options:[NSDictionary dictionary]];
	[self.bottomMarginField bind:NSValueBinding toObject:self withKeyPath:PageLayoutBottomMarginKey options:[NSDictionary dictionary]];
	[self.unitsPopUp removeAllItems];
	[self.unitsPopUp addItemsWithTitles:[[self pointsPerUnit] allKeys]];
	if (YES == [defaultUnit isEqualToString:@"Metric"])
		[self.unitsPopUp selectItemWithTitle:@"cm"];
	else {
		if ((YES == [defaultUnit isEqualToString:@"U.S."]) || (YES == [defaultUnit isEqualToString:@"Inches"]))
			[self.unitsPopUp selectItemWithTitle:@"inch"];
		else {
			if (YES == [defaultUnit isEqualToString:@"Picas"])
				[self.unitsPopUp selectItemWithTitle:@"pica"];
			else
				[self.unitsPopUp selectItemWithTitle:@"point"];
		}
	}
	self.leftMarginField.cell.sendsActionOnEndEditing = YES;
	self.rightMarginField.cell.sendsActionOnEndEditing = YES;
	self.topMarginField.cell.sendsActionOnEndEditing = YES;
	self.bottomMarginField.cell.sendsActionOnEndEditing = YES;
	// horizontalPaginationPopUp, verticalPaginationPopUpのindexは0=NSAutoPagination,1=NSFitPagination,2=NSClipPaginationとする。 /System/Library/Frameworks/AppKit.framework/NSPrintInfo.hによる。 //
	// horizontalPaginationPopUpとverticalPaginationPopUpのindexは0:Diced 1:Scaled 2:Cliped でなければならない。printInfo.horizontallyCenteredとprintInfo.verticallyCenteredとで設定する番号に一致させる。IBとASPageLayoutAccessory.stringsとで設定する。 //
	self.originView.titleAlignment = NSTextAlignmentRight; // InterfaceBuilderにおいて配列の順番を与えるtagを設定しておかなければならない。 //
	self.originView.textAlignment = NSTextAlignmentLeft;
	// [self.originView calcSize]; としてもASFormの設定が反映されない。 //
	[[self.originView cellAtIndex:0] bind:NSValueBinding toObject:self withKeyPath:PageLayoutImageableAreaOriginXKey options:[NSDictionary dictionary]];
	[[self.originView cellAtIndex:1] bind:NSValueBinding toObject:self withKeyPath:PageLayoutImageableAreaOriginYKey options:[NSDictionary dictionary]];
	self.widthView.titleAlignment = NSTextAlignmentRight;
	self.widthView.textAlignment = NSTextAlignmentLeft;
	[[self.widthView cellAtIndex:0] bind:NSValueBinding toObject:self withKeyPath:PageLayoutLeftSpaceKey options:[NSDictionary dictionary]];
	[[self.widthView cellAtIndex:1] bind:NSValueBinding toObject:self withKeyPath:PageLayoutRightSpaceKey options:[NSDictionary dictionary]];
	[self.heightView setTitleAlignment:NSTextAlignmentRight];
	[self.heightView setTextAlignment:NSTextAlignmentLeft];
	[[self.heightView cellAtIndex:0] bind:NSValueBinding toObject:self withKeyPath:PageLayoutTopSpaceKey options:[NSDictionary dictionary]];
	[[self.heightView cellAtIndex:1] bind:NSValueBinding toObject:self withKeyPath:PageLayoutBottomSpaceKey options:[NSDictionary dictionary]];
	NSNumberFormatter *sizeFormatter = [[NSNumberFormatter alloc] init];
	NSString *sizeFormat = [NSString stringWithFormat:@"#,##0.000;0.000;👎-#,##0.000"]; // 👎(THUMBS DOWN SIGN)は\Ud83d\Udc4eであるが直接書いても良いし、フォントの指定(Apple Color Emoji)もしなくて良い。ASPageLayoutAccessory.stringsはProprty List Textとして"Key"="Value"の記述ができて、NSLocalizedStringFromTableInBundle(Key,@"ASPageLayoutAccessory",[NSBundle bundleForClass:self.class],Value)が使える。 //
	sizeFormatter.formatterBehavior = NSNumberFormatterBehavior10_4;
	sizeFormatter.format = sizeFormat;
	self.leftMarginField.cell.formatter = sizeFormatter;
	self.rightMarginField.cell.formatter = sizeFormatter;
	self.topMarginField.cell.formatter = sizeFormatter;
	self.bottomMarginField.cell.formatter = sizeFormatter;
	[self.originView cellAtIndex:0].formatter = sizeFormatter;
	[self.originView cellAtIndex:1].formatter = sizeFormatter;
	[self.widthView cellAtIndex:0].formatter = sizeFormatter;
	[self.widthView cellAtIndex:1].formatter = sizeFormatter;
	[self.heightView cellAtIndex:0].formatter = sizeFormatter;
	[self.heightView cellAtIndex:1].formatter = sizeFormatter;
	[sizeFormatter release];
	[self changeUnits:self.unitsPopUp];
}

- (void)setRepresentedObject:(id)printInfo {
	// printInfoがnilになることはないようだ。 //
	// [NSPrintInfo setTopMargin:]で用紙の下余白が設定される。NSMinY(imageablePageBounds)の値に関係なく余白が設定される。[NSPrintInfo setBottomMargin:]で用紙の上余白が設定される。NSMaxY(imageablePageBounds)の値に関係なく余白が設定される。PageLayoutTopSpaceKeyはimageablePageBoundsの中での余裕の値。負数のときは下側の画像が一部切り取られることになる。 //
	// [NSPrintInfo setLeftMargin:]で用紙の左余白が設定される。NSMinX(imageablePageBounds)の値に関係なく余白が設定される。負数のときNSMinX(imageablePageBounds)の位置から描画され、右側の方に負数の分だけシフトされるので1ページに収まらなくなる。[NSPrintInfo setRightMargin:]で用紙の右余白が設定される。負数のとき右側に拡張されるので1ページに収まらない。上下左右とも負数になるかどうかを見るためにPageLayoutLeftSpaceKeyなどを表示させる。 //
	// 以上から判るように用紙の上下左右の余白は設定したマージンによってのみ定まり、プリンターの機種の相違によるimageablePageBoundsの差異は吸収されて同一になる。 //
	[super setRepresentedObject:printInfo];
	if (nil != printInfo) {
		NSNumber *hAndFNumber = [[printInfo dictionary] objectForKey:NSPrintHeaderAndFooter]; // printInfoがnilでなくてもNSPrintHeaderAndFooterの値がnilのときがある。 //
		BOOL hAndF = (nil != hAndFNumber) ? hAndFNumber.boolValue : NO;
		NSRect imageable = [printInfo imageablePageBounds];
		[self setValue:[NSNumber numberWithDouble:[printInfo leftMargin]] forKey:PageLayoutLeftMarginKey];
		[self setValue:[NSNumber numberWithDouble:[printInfo rightMargin]] forKey:PageLayoutRightMarginKey];
		[self setValue:[NSNumber numberWithDouble:NSMinX(imageable)] forKey:PageLayoutImageableAreaOriginXKey];
		[self setValue:[NSNumber numberWithDouble:NSMinY(imageable)] forKey:PageLayoutImageableAreaOriginYKey];
		[self setValue:[NSNumber numberWithDouble:[printInfo leftMargin] - NSMinX(imageable)] forKey:PageLayoutLeftSpaceKey];
		[self setValue:[NSNumber numberWithDouble:[printInfo rightMargin] - [printInfo paperSize].width + NSMaxX(imageable)] forKey:PageLayoutRightSpaceKey];
		if (YES == self.upsideDown) {
			[self setValue:[NSNumber numberWithDouble:[printInfo bottomMargin]] forKey:PageLayoutTopMarginKey]; // 上下反転 //
			[self setValue:[NSNumber numberWithDouble:[printInfo topMargin]] forKey:PageLayoutBottomMarginKey]; // 上下反転 //
			[self setValue:[NSNumber numberWithDouble:[printInfo bottomMargin] - [printInfo paperSize].height + NSMaxY(imageable)] forKey:PageLayoutTopSpaceKey]; // 上下反転 //
			[self setValue:[NSNumber numberWithDouble:[printInfo topMargin] - NSMinY(imageable)] forKey:PageLayoutBottomSpaceKey]; // 上下反転 //
		}
		else {
			[self setValue:[NSNumber numberWithDouble:[printInfo topMargin]] forKey:PageLayoutTopMarginKey];
			[self setValue:[NSNumber numberWithDouble:[printInfo bottomMargin]] forKey:PageLayoutBottomMarginKey];
			[self setValue:[NSNumber numberWithDouble:[printInfo topMargin] - [printInfo paperSize].height + NSMaxY(imageable)] forKey:PageLayoutTopSpaceKey];
			[self setValue:[NSNumber numberWithDouble:[printInfo bottomMargin] - NSMinY(imageable)] forKey:PageLayoutBottomSpaceKey];
		}
		[self setValue:[NSNumber numberWithBool:hAndF] forKey:PageLayoutHeaderAndFooterKey];
		[self setValue:[NSNumber numberWithBool:[printInfo isVerticallyCentered]] forKey:PageLayoutVerticallyCenterdKey];
		[self setValue:[NSNumber numberWithBool:[printInfo isHorizontallyCentered]] forKey:PageLayoutHorizontallyCenterdKey];
		[self setValue:[NSNumber numberWithInteger:[printInfo verticalPagination]] forKey:PageLayoutVerticalPaginationKey]; // /System/Library/Frameworks/AppKit.framework/NSPrintInfo.h  NSAutoPagination=0,NSFitPagination=1,NSClipPagination=2 //
		[self setValue:[NSNumber numberWithInteger:[printInfo horizontalPagination]] forKey:PageLayoutHorizontalPaginationKey];
	}
	else
		NSLog(@"printInfo is nil in ASPageLayoutAccessory.");
	self.headerAndFooterSwitch.enabled = self.headerAndFooterEnabled;
}

- (void)setLeftMargin:(CGFloat)aNumber {
	NSPrintInfo *printInfo = self.representedObject;
	NSRect imageable = printInfo.imageablePageBounds;
	self._leftMargin = aNumber;
	[self setValue:[NSNumber numberWithDouble:aNumber - NSMinX(imageable)] forKey:PageLayoutLeftSpaceKey];
	printInfo.leftMargin = aNumber;
}
- (void)setRightMargin:(CGFloat)aNumber {
	NSPrintInfo *printInfo = self.representedObject;
	NSRect imageable = printInfo.imageablePageBounds;
	self._rightMargin = aNumber;
	[self setValue:[NSNumber numberWithDouble:aNumber - printInfo.paperSize.width + NSMaxX(imageable)] forKey:PageLayoutRightSpaceKey];
	printInfo.rightMargin = aNumber;
}
- (void)setTopMargin:(CGFloat)aNumber {
	NSPrintInfo *printInfo = self.representedObject;
	NSRect imageable = printInfo.imageablePageBounds;
	self._topMargin = aNumber;
	[self setValue:[NSNumber numberWithDouble:aNumber - printInfo.paperSize.height + NSMaxY(imageable)] forKey:PageLayoutTopSpaceKey];
	if (YES == self.upsideDown)
		printInfo.bottomMargin = aNumber; // 上下反転 //
	else
		printInfo.topMargin = aNumber;
}
- (void)setBottomMargin:(CGFloat)aNumber {
	NSPrintInfo *printInfo = self.representedObject;
	NSRect imageable = printInfo.imageablePageBounds;
	self._bottomMargin = aNumber;
	[self setValue:[NSNumber numberWithDouble:aNumber - NSMinY(imageable)] forKey:PageLayoutBottomSpaceKey];
	if (YES == self.upsideDown)
		printInfo.topMargin = aNumber; // 上下反転 //
	else
		printInfo.bottomMargin = aNumber;
}
- (void)setHeaderAndFooter:(BOOL)flag {
    NSPrintInfo *printInfo = self.representedObject;
	self._headerAndFooter = flag;
    [printInfo.dictionary setObject:[NSNumber numberWithBool:flag] forKey:NSPrintHeaderAndFooter];
}
- (void)setVerticallyCentered:(BOOL)flag {
    NSPrintInfo *printInfo = self.representedObject;
	self._verticallyCentered = flag;
	[self.centeredMatrix selectCellAtRow:(NO == flag) ? 0 : 1 column:(NO == flag) ? 0 : 1];
	printInfo.verticallyCentered = flag;
}
- (void)setHorizontallyCentered:(BOOL)flag {
    NSPrintInfo *printInfo = self.representedObject;
	self._horizontallyCentered = flag;
	[self.centeredMatrix selectCellAtRow:(NO == flag) ? 0 : 1 column:(NO == flag) ? 0 : 1];
	printInfo.horizontallyCentered = flag;
}
- (void)setVerticalPagination:(NSInteger)aNumber {
    NSPrintInfo *printInfo = self.representedObject;
	self._verticalPagination = aNumber;
	printInfo.verticalPagination = aNumber;
}
- (void)setHorizontalPagination:(NSInteger)aNumber {
    NSPrintInfo *printInfo = self.representedObject;
	self._horizontalPagination = aNumber;
	printInfo.horizontalPagination = aNumber;
}

- (IBAction)changeCenterd:(id)sender {
	NSInteger row = [sender selectedRow];
	NSInteger column = [sender selectedColumn];
	[self setValue:[NSNumber numberWithBool:(0 == row) ? NO : YES] forKey:PageLayoutVerticallyCenterdKey];
	[self setValue:[NSNumber numberWithBool:(0 == column) ? NO : YES] forKey:PageLayoutHorizontallyCenterdKey];
}
- (IBAction)changeUnits:(id)sender {
	NSString *units = [sender titleOfSelectedItem];
	NSNumber *ppu = [[self pointsPerUnit] objectForKey:units];
	NSNumber *_ppu = [NSNumber numberWithDouble:1.0 / ppu.doubleValue];
	NSNumberFormatter *numberFormatter = self.leftMarginField.cell.formatter;

	numberFormatter.multiplier = _ppu;
	numberFormatter = self.rightMarginField.cell.formatter;
	numberFormatter.multiplier = _ppu;
	numberFormatter = self.topMarginField.cell.formatter;
	numberFormatter.multiplier = _ppu;
	numberFormatter = self.bottomMarginField.cell.formatter;
	numberFormatter.multiplier = _ppu;
	numberFormatter = [[self.originView cellAtIndex:0] formatter];
	numberFormatter.multiplier = _ppu;
	numberFormatter = [[self.originView cellAtIndex:1] formatter];
	numberFormatter.multiplier = _ppu;
	numberFormatter = [[self.widthView cellAtIndex:0] formatter];
	numberFormatter.multiplier = _ppu;
	numberFormatter = [[self.widthView cellAtIndex:1] formatter];
	numberFormatter.multiplier = _ppu;
	numberFormatter = [[self.heightView cellAtIndex:0] formatter];
	numberFormatter.multiplier = _ppu;
	numberFormatter = [[self.heightView cellAtIndex:1] formatter];
	numberFormatter.multiplier = _ppu;
	NSPrintInfo *printInfo = self.representedObject;
	// macOS10.15でsetFloatValue:が必要になった。 //
	self.leftMarginField.cell.floatValue = printInfo.leftMargin;
	self.rightMarginField.cell.floatValue = printInfo.rightMargin;
	self.topMarginField.cell.floatValue = printInfo.topMargin;
	self.bottomMarginField.cell.floatValue = printInfo.bottomMargin;
	self.leftMarginField.needsDisplay = YES;
	self.rightMarginField.needsDisplay = YES;
	self.topMarginField.needsDisplay = YES;
	self.bottomMarginField.needsDisplay = YES;
	[[self.originView cellAtIndex:0] setNeedsDisplay:YES];
	[[self.originView cellAtIndex:1] setNeedsDisplay:YES];
	[[self.widthView cellAtIndex:0] setNeedsDisplay:YES];
	[[self.widthView cellAtIndex:1] setNeedsDisplay:YES];
	[[self.heightView cellAtIndex:0] setNeedsDisplay:YES];
	[[self.heightView cellAtIndex:1] setNeedsDisplay:YES];
}

- (void)cursorUpdate:(NSEvent *)event {
	// cursorUpdate:を使うのには意味が無い。ただ画面に表示されるときに呼び出されるからである。 //
	[super cursorUpdate:event];
	NSPrintInfo *printInfo = self.representedObject;
	NSRect imageable = printInfo.imageablePageBounds;
	
	[self setValue:[NSNumber numberWithDouble:NSMinX(imageable)] forKey:PageLayoutImageableAreaOriginXKey];
	[self setValue:[NSNumber numberWithDouble:NSMinY(imageable)] forKey:PageLayoutImageableAreaOriginYKey];
	[self setValue:[NSNumber numberWithDouble:printInfo.leftMargin - NSMinX(imageable)] forKey:PageLayoutLeftSpaceKey];
	[self setValue:[NSNumber numberWithDouble:printInfo.rightMargin - printInfo.paperSize.width + NSMaxX(imageable)] forKey:PageLayoutRightSpaceKey];
	if (YES == self.upsideDown) {
		[self setValue:[NSNumber numberWithDouble:printInfo.bottomMargin - printInfo.paperSize.height + NSMaxY(imageable)] forKey:PageLayoutTopSpaceKey]; // 上下反転 //
		[self setValue:[NSNumber numberWithDouble:printInfo.topMargin - NSMinY(imageable)] forKey:PageLayoutBottomSpaceKey]; // 上下反転 //
	}
	else {
		[self setValue:[NSNumber numberWithDouble:printInfo.topMargin - printInfo.paperSize.height + NSMaxY(imageable)] forKey:PageLayoutTopSpaceKey];
		[self setValue:[NSNumber numberWithDouble:printInfo.bottomMargin - NSMinY(imageable)] forKey:PageLayoutBottomSpaceKey];
	}
	[self.originView calcSize]; // ここでやるしかない? //
	[self.widthView calcSize];
	[self.heightView calcSize];
}

@end
